# Copyright 3D Control Systems, Inc. All Rights Reserved 2017-2019. Built in San Francisco.
#
# This software is distributed under commercial non-GPL license for personal, educational,
# corporate or any other use. The software as a whole or any parts of that are prohibited
# for distribution and/or use without obtaining license from 3D Control Systems, Inc.
#
# If you do not have the license to use this software, please delete all software files
# immediately and contact sales to obtain the license: sales@3dprinteros.com.
# If you are unsure about the licensing please contact directly our sales: sales@3dprinteros.com.


import time
import logging
import threading
import urllib.request
from PySide2 import QtCore

import log

try:
    NM = None
    import NetworkManager as NM
except ImportError:
    logging.getLogger("NetworkModel").warning("Error importing python-networkmanager: no module")
except:
    logging.getLogger("NetworkModel").warning("Error importing python-networkmanager: cant connect to networkmanager. Is it up?")


def wifi_strength_percent_to_level(strength_percent):
    return round(strength_percent / 25)


class NetworkModel(QtCore.QObject):

    UPDATE_LOOP_TIME = 1
    AFTER_SCAN_SLEEP = 12
    NETWORK_CHECK_TIMEOUT = 5
    CONNECTION_WAIT_TIMEOUT = 5
    NETWORK_CHECK_URL = 'https://cloud.3dprinteros.com/noauth/get_timestamp'

    on_connectConnected = QtCore.Signal()
    on_networkConnectChanged = QtCore.Signal()
    on_connectError = QtCore.Signal()
    on_wifiListChanged = QtCore.Signal()
    on_wifiSignal = QtCore.Signal()
    on_localIp = QtCore.Signal()

    def __init__(self, core_model):
        super().__init__()
        self.logger = logging.getLogger(__name__)
        self._coreModel = core_model
        self._app = core_model._app
        if NM:
            self._networkManager = NM.NetworkManager
        self._localIp = "N/A"
        self._wifiStationList = []
        self._wifiNMAPObjsDict = {}
        self._wifiScanLock = threading.RLock()
        self._wifiInterface = self._initWifiInterface()
        self._networkConnected = True
        self._networkConnected = self._checkNetwork()
        self._connectError = ""
        self._connectThread = None
        self._networkCheckThread = threading.Thread(target=self._networkCheckLoop)
        self._networkCheckThread.start()

    def _checkNetwork(self):
        try:
            if NM:
                state = self._networkManager.State
                check = state in (NM.NM_STATE_CONNECTED_GLOBAL, NM.NM_STATE_CONNECTED_LOCAL, NM.NM_STATE_CONNECTED_SITE)
                if check:
                    try:
                        connection = self._networkManager.ActiveConnections[0]
                        self._localIp = connection.Ip4Config.Addresses[0][0]
                    except (TypeError, AttributeError):
                        self._localIp = "N/A"
                else:
                    self._localIp = "N/A"
            else:
                check = bool(urllib.request.urlopen(self.NETWORK_CHECK_URL, timeout=self.NETWORK_CHECK_TIMEOUT))
        except:
            check = False
        if self._networkConnected != check:
            self._networkConnected = check
            self.on_networkConnectChanged.emit()
            if check:
                self.logger.warning("Network connection lost")
            else:
                self.logger.warning("Network connection entablished")
        return check

    def _checkSignalLevel(self):
        signal_level = 0
        try:
            if self._wifiInterface:
                if self._wifiInterface.ActiveAccessPoint:
                    signal_level = wifi_strength_percent_to_level(self._wifiInterface.ActiveAccessPoint.Strength)
        except:
            self.logger.warning("Error getting WiFi signal")
        return signal_level

    @log.log_exception
    def _networkCheckLoop(self):
        self._wifiSignal = 0
        while not self._app.stop_flag:
            self._checkNetwork()
            new_signal_level = self._checkSignalLevel()
            if new_signal_level != self._wifiSignal:
                self._wifiSignal = new_signal_level
                self.on_wifiSignal.emit()
                self.on_wifiListChanged.emit()
            time.sleep(self.UPDATE_LOOP_TIME)

    def _initWifiInterface(self):
        if NM:
            for device in self._networkManager.GetAllDevices():
                if device.DeviceType == NM.NM_DEVICE_TYPE_WIFI:
                    self.logger.info("Got wifi device: " + str(device.Interface))
                    self._wifiInterface = device
                    break

    def _getWiFiInterface(self):
        try:
            interfaces = self._pywifi.interfaces()
        except OSError:
            self.logger.warning("Warning: error getting WiFi interfaces list")
        else:
            for interface in interfaces:
                name = interface.name()
                if name in self._correctInterfaceNames:
                    self._wifiInterfaceName = name
                    self._wifiInterface = interface
                    self.logger.info("Got wifi interface " + name)
                    return True
            else:
                self.logger.warning("Warning: error getting WiFi interface")
                return False

    def _connectWifi(self, ssid, password):
        self.logger.debug("Connecting to ssid: " + str(ssid))
        self._connectError == "" # clean prev error
        with self._wifiScanLock:
            self.scanWifi()
            access_point_object = self._wifiNMAPObjsDict.get(ssid)
            if not access_point_object:
                message = "No access point with such ssid: " + str(ssid)
                self.logger.exception(message)
                self._connectError = message
                self.on_connectError.emit()
                return False
            settings = {'802-11-wireless-security': {'psk': password}}
            try:
                self._networkManager.AddAndActivateConnection(settings, self._wifiInterface, access_point_object)
            except: #TODO add appropriate dbus excetion here
                message = "Error connecting to WiFi"
                self.logger.exception(message)
                self._connectError = message
                self.on_connectError.emit()
            else:
                while not self._app.stop_flag:
                    time.sleep(1)
                    state = self._networkManager.State
                    if state == NM.NM_STATE_CONNECTING:
                        continue
                    if state == NM.NM_STATE_DISCONNECTED:
                        self._connectError = "Error connecting to WiFi. Wrong password?"
                        self.on_connectError.emit()
                        return False
                    if state in (NM.NM_STATE_CONNECTED_GLOBAL, NM.NM_STATE_CONNECTED_LOCAL, NM.NM_STATE_CONNECTED_SITE):
                        self.on_connectConnected.emit()
                        return True

    def _disconnectWiFi(self):
        self.logger.info("Disconnecting from WiFi")
        try:
            active_connection = None
            if self._wifiInterface:
                active_connection = self._wifiInterface.ActiveConnection
            if active_connection:
                self._wifiInterface.Disconnect()
                self.logger.info("WiFi disconnect success")
            else:
                self.logger.info("No connection to disconnect from. Skipping.")
        except Exception as e:
            self.logger.warning("Error while disconnecting WiFi:" + str(e))
            self.scanWifi()

    @QtCore.Slot(str)
    def forgetNetwork(self, ssid):
        self.logger.error("Error: forgetNetwork is not implemented yet")

    @QtCore.Slot()
    def scanWifi(self):
        self.logger.debug("Scanning...")
        with self._wifiScanLock:
            if not self._wifiInterface:
                self._initWifiInterface()
            connected_ssid = None
            try:
                connected_access_point = self._wifiInterface.ActiveAccessPoint
            except:
                self.logger.warning("Error getting current access point")
            else:
                if connected_access_point:
                    connected_ssid = connected_access_point.Ssid
            try:
                all_access_points = self._wifiInterface.GetAllAccessPoints()
            except:
                self.logger.error("Error getting access points")
                return False
            if len(all_access_points) < 2:
                self.logger.warning("Warning: suspisiosly short wifi access points list. Wrong mode?")
            wifiStationList = []
            wifiNMAPObjsDict = {}
            for access_point in all_access_points:
                try:
                    ssid = access_point.Ssid
                    signalLevel = wifi_strength_percent_to_level(access_point.Strength)
                    encrypted = not access_point.Flags & NM.NM_802_11_AP_FLAGS_NONE
                    #infrastusture_mode = bool(access_point.Mode & NM.NM_802_11_MODE_INFRA)
                    #adhog_mode = bool(access_point.Mode & NM.NM_802_11_MODE_ADHOC)
                except:
                    self.logger.exception("Access point vanished")
                else:
                    self.logger.info("WiFi scan found: " + ssid)
                    access_point_dict = {"ssid": ssid,
                                    "secured": encrypted,
                                    "signalLevel": signalLevel,
                                    #"infrastusture_mode": infrastusture_mode,
                                    #"adhog_mode": adhog_mode,
                                    "connected": ssid == connected_ssid}
                    wifiStationList.append(access_point_dict)
                    wifiNMAPObjsDict[ssid] = access_point
            self._wifiStationList = wifiStationList
            self._wifiNMAPObjsDict = wifiNMAPObjsDict
            self.on_wifiListChanged.emit()
            #self.logger.debug(self._wifiStationList)
            return True

    @QtCore.Slot(str, str)
    def connectWifi(self, ssid, passwd = ""):
        self.logger.debug("Connecting Wi-Fi %s:%s" % (ssid, passwd))
        if self._connectThread and self._connectThread.is_alive():
            self.logger.error("Error: another connection is already in progress")
        else:
            self._connectThread = threading.Thread(target=self._connectWifi, args=(ssid, passwd))
            self._connectThread.start()

    @QtCore.Slot()
    def disconnectWifi(self):
        self._disconnectWiFi()

    @QtCore.Property(int, notify=on_wifiSignal)
    def wifiSignal(self):
        return self._wifiSignal

    @QtCore.Property(str, notify=on_connectError)
    def connectError(self):
        return self._connectError

    @QtCore.Property('QVariantList', notify=on_wifiListChanged)
    def wifiStationList(self):
        return self._wifiStationList

    @QtCore.Property(bool, notify=on_networkConnectChanged)
    def networkConnected(self):
        return self._networkConnected

    @QtCore.Property(str, notify=on_localIp)
    def localIp(self):
        return self._localIp
